from typing import Optional, Dict, Any, List, Union, ClassVar
import logging
import websockets
from inspect import isclass
from concurrent.futures import ThreadPoolExecutor
from rpc_gateway import errors, messages, gateway
from rpc_gateway.utils import await_sync

logger = logging.getLogger(__name__)


class Server(gateway.GatewayClient):
    def __init__(self, gateway_url: str = 'ws://localhost:8888', max_workers: int = 5, instances: Optional[Dict[str, Any]] = None):
        super().__init__(gateway_url)
        self.instances = {} if instances is None else instances
        self.executor = ThreadPoolExecutor(max_workers=max_workers)

    async def _on_start(self):
        await self._register_gateway_instances(list(self.instances.keys()))

    async def _register_gateway_instances(self, instances: List[str]):
        await self.websocket_connection.send_request(messages.Method.REGISTER, instances)

    def register(self, name: str, instance: Any) -> Any:
        if name in self.instances:
            raise errors.InstanceAlreadyRegisteredError(f'Instance already registered with name: {name}')

        self.instances[name] = instance

        return instance

    def _get_instance(self, instance_name: Union[type, str]):
        try:
            if isclass(instance_name):
                return self.instances[instance_name.__name__]

            return self.instances[instance_name]
        except KeyError:
            raise errors.InstanceNotFoundError(f'Instance not found: {instance_name}')

    def _get(self, instance_name: str, attribute_name: str) -> messages.Response:
        try:
            instance = self._get_instance(instance_name)
            data = getattr(instance, attribute_name)

            if callable(data):
                return messages.Response(status=messages.Status.METHOD)

            return messages.Response(data=data)
        except Exception as err:
            return messages.Response(status=messages.Status.ERROR, data=err)

    def _set(self, instance_name: str, attribute_name: str, value: Any) -> messages.Response:
        try:
            instance = self._get_instance(instance_name)
            setattr(instance, attribute_name, value)
            return messages.Response()
        except Exception as err:
            return messages.Response(status=messages.Status.ERROR, data=err)

    def _call(self, instance_name: str, attribute_name: str, args: List[Any], kwargs: Dict[str, Any]) -> messages.Response:
        try:
            instance = self._get_instance(instance_name)
            method = getattr(instance, attribute_name)
            data = method(*args, **kwargs)
            return messages.Response(data=data)
        except Exception as err:
            return messages.Response(status=messages.Status.ERROR, data=err)

    async def _run(self, *args: Any) -> Any:
        return await self.event_loop.run_in_executor(self.executor, *args)

    async def _handle_request(self, websocket_connection: websockets.WebSocketCommonProtocol, request: messages.Request) -> messages.Response:
        if request.method == 'get':
            return await self._run(self._get, request.data['instance'], request.data['attribute'])

        if request.method == 'set':
            return await self._run(self._set, request.data['instance'], request.data['attribute'], request.data['value'])

        if request.method == 'call':
            return await self._run(self._call, request.data['instance'], request.data['attribute'], request.data['args'], request.data['kwargs'])

        return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMethodError(f'Invalid method: {request.method}'))


if __name__ == '__main__':
    import time
    logging.basicConfig(level=logging.INFO)

    class TestClass:
        foo = 'bar'

        def method(self):
            return 'baz'

        def sleep(self, duration):
            time.sleep(duration)

    server = Server()
    server.register('test', TestClass())
    server.start()


from typing import Any, Dict, Callable, Optional, Union, List, Tuple
from inspect import isclass
from os import path
import json
import yaml
from deepmerge import always_merger

FactoryType = Callable[..., Any]
ConfigurationType = Dict[str, Union[List, Dict[str, Any]]]
GetItemType = Union[str, FactoryType]


def configure(config: Union[str, ConfigurationType], file_type: Optional[str] = None):
    Registry.default_registry().configure(config, file_type)


def get(item: GetItemType, *args: Any, **kwargs: Any) -> Any:
    return Registry.default_registry().get(item, *args, **kwargs)


def register(factory: FactoryType, name: Optional[str] = None):
    Registry.default_registry().register(factory, name)


def override_factory(item: GetItemType, factory: FactoryType):
    Registry.default_registry().override_factory(item, factory)


def default_registry() -> 'Registry':
    return Registry.default_registry()


class Registry:
    CONFIG_META_PREFIX = '$'
    ARGS_KEY = '$args'
    TYPE_KEY = '$type'
    _default_registry: Optional['Registry'] = None

    @staticmethod
    def reset():
        Registry._default_registry = None

    @staticmethod
    def default_registry() -> 'Registry':
        if Registry._default_registry is None:
            Registry._default_registry = Registry()

        return Registry._default_registry

    def __init__(self,
                 config_path: Optional[str] = None,
                 configuration: Optional[ConfigurationType] = None,
                 factories: Optional[Dict[str, FactoryType]] = None,
                 objects: Optional[Dict[str, Any]] = None,
                 set_default: bool = True):
        self.config_path = config_path
        self.configuration = configuration or {}
        self.factories = factories or {}
        self.factory_types = {val: key for key, val in self.factories.items()}
        self.objects = objects or {}

        if set_default and Registry._default_registry is None:
            Registry._default_registry = self

        if self.config_path is not None:
            self.load_configuration()

    def reset_object(self, object_name: str):
        self.objects.pop(object_name)

    def configure(self, config: Union[str, ConfigurationType], file_type: Optional[str] = None):
        if isinstance(config, dict):
            self.merge_configuration(config)
        else:
            self.load_configuration(config, file_type)

    def merge_configuration(self, config: ConfigurationType):
        self.configuration = always_merger.merge(self.configuration or {}, config)

    def load_configuration(self, config_path: Optional[str] = None, file_type: Optional[str] = None):
        file_path = config_path or self.config_path

        # use the file extension as the file type if none is given
        if file_type is None:
            _, config_ext = path.splitext(file_path)
            config_type = config_ext[1:]  # remove .
        else:
            config_type = file_type

        with open(file_path or self.config_path) as config_file:
            if config_type == 'json':
                self.merge_configuration(json.load(config_file))
            elif config_type == 'yaml':
                self.merge_configuration(yaml.safe_load(config_file))
            else:
                raise RegistryError(f'Invalid config file format, must be "json" or "yaml": {config_type}')

    def get_object_meta_configuration(self, object_name: str) -> Dict[str, Any]:
        if object_name not in self.configuration:
            return {}

        config = self.configuration[object_name]
        # remove any keys that don't start with @ (and aren't @args) and remove the @ from the keys
        return {k[1:]: v for k, v in config.items() if k.startswith(self.CONFIG_META_PREFIX) and k != self.ARGS_KEY}

    def get_object_configuration(self, object_name: Optional[str], *args: Any, **kwargs: Any) -> Tuple[List, Dict]:
        if object_name is None or object_name not in self.configuration:
            return [], {}

        config = self.configuration[object_name]

        if isinstance(config, list):
            return [*config, *args], kwargs

        # start with the config of the factory defined by the $type key in the config, if it exists
        config_args, config_kwargs = self.get_object_configuration(config.get(self.TYPE_KEY, None))

        # merge object config, if defined
        config_args = [*config_args, *config.get(self.ARGS_KEY, [])]
        config_kwargs = {**config_kwargs, **{k: v for k, v in config.items() if not k.startswith(self.CONFIG_META_PREFIX)}}

        # finally merge in any additional args and kwargs
        return [*config_args, *args], {**config_kwargs, **kwargs}

    def get_object_factory(self, object_name: str) -> FactoryType:
        if object_name not in self.factories:
            if object_name not in self.configuration:
                raise RegistryFactoryNotFoundError(f'Unable to find a factory or config for object: {object_name}')

            if self.TYPE_KEY not in self.configuration[object_name]:
                raise RegistryFactoryNotFoundError(
                    f'No factory or "$type" key found in config for object: {object_name}')

            factory_name = self.configuration[object_name][self.TYPE_KEY]

            if factory_name not in self.configuration:
                raise RegistryFactoryNotFoundError(
                    f'Unable to find a factory for "{object_name}" with $type: {factory_name}')

            return self.factories[factory_name]

        return self.factories[object_name]

    def register(self, factory: FactoryType, name: Optional[str] = None):
        if name is None:
            if not isclass(factory):
                raise RegistryError(f'Only classes can be registered without a name')

            name = factory.__name__

        if name in self.factories:
            raise RegistryFactoryExists(f'Factory with name "{name}" already registered')

        self.factories[name] = factory
        self.factory_types[factory] = self.factory_types.get(factory, []) + [name]

    def override_factory(self, item: GetItemType, factory: FactoryType):
        object_name = item if isinstance(item, str) else item.__name__

        if object_name not in self.factories:
            raise RegistryFactoryNotFoundError(f'Cannot override factory, existing factory not found for item: {item}')

        self.factories[object_name] = factory

    def create_object(self, object_name: str, *args: Any, **kwargs: Any):
        factory = self.get_object_factory(object_name)
        object_args, object_kwargs = self.get_object_configuration(object_name, *args, **kwargs)
        return factory(*object_args, **object_kwargs)

    def get_by_name(self, object_name: str, *args: Any, **kwargs: Any):
        if object_name in self.objects:
            return self.objects[object_name]

        self.objects[object_name] = self.create_object(object_name, *args, **kwargs)
        return self.objects[object_name]

    def get_by_type(self, factory: FactoryType, *args: Any, **kwargs: Any) -> Any:
        if factory not in self.factory_types:
            raise RegistryFactoryNotFoundError(f'Unable to find factory: {factory}')

        if not isclass(factory):
            raise RegistryFactoryIsNotClassError(f'Cannot get default object, factory must be a class: {factory}')

        object_names = self.factory_types[factory]
        object_name = factory.__name__

        if object_name not in object_names:
            raise RegistryFactoryDefaultNotFoundError(f'Unable to find a default object for factory: {factory}')

        return self.get_by_name(object_name, *args, **kwargs)

    def get(self, item: GetItemType, *args: Any, **kwargs: Any) -> Any:
        if isinstance(item, str):
            return self.get_by_name(item, *args, **kwargs)

        return self.get_by_type(item, *args, **kwargs)


class RegistryError(Exception):
    pass


class RegistryFactoryNotFoundError(RegistryError):
    pass


class RegistryFactoryDefaultNotFoundError(RegistryError):
    pass


class RegistryFactoryExists(RegistryError):
    pass


class RegistryFactoryIsNotClassError(RegistryError):
    pass



from typing import Any, List, Dict
from dataclasses import dataclass, asdict
import json
import pickle


class Status:
    SUCCESS = 0
    ERROR = 1
    METHOD = 2


class Method:
    GET = 'get'
    SET = 'set'
    CALL = 'call'
    REGISTER = 'register'
    LOCK = 'lock'
    UNLOCK = 'unlock'
    AVAILABLE = 'available'


@dataclass
class Message:
    id: int = 0
    data: Any = None

    @classmethod
    def from_json(cls, json_str: str) -> 'Message':
        return cls(**json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    def to_pickle(self) -> bytes:
        return pickle.dumps(self)


@dataclass
class Response(Message):
    status: int = Status.SUCCESS


@dataclass
class Request(Message):
    method: str = Method.GET



class HardwareGatewayError(Exception):
    pass


class InstanceNotFoundError(HardwareGatewayError):
    pass


class InvalidMethodError(HardwareGatewayError):
    pass


class InstanceAlreadyRegisteredError(HardwareGatewayError):
    pass


class InvalidMessageError(HardwareGatewayError):
    pass


class InvalidRequestError(HardwareGatewayError):
    pass


class InvalidResponseError(HardwareGatewayError):
    pass


class InvalidPathError(HardwareGatewayError):
    pass


class NotConnectedError(HardwareGatewayError):
    pass


class InvalidMessageIdError(HardwareGatewayError):
    pass


class RequestTimeoutError(HardwareGatewayError):
    pass


class ConnectionError(HardwareGatewayError):
    pass


class ServerConnectionLostError(HardwareGatewayError):
    pass


class SerializationError(HardwareGatewayError):
    pass



from typing import Optional, Dict, Any, Callable, Awaitable
import asyncio
import pickle
import logging
import websockets
from concurrent import futures
from rpc_gateway import messages, errors, utils

logger = logging.getLogger(__name__)
RequestHandlerType = Callable[['WebsocketConnection', messages.Request], Awaitable[messages.Response]]
CloseHandlerType = Callable[['MessagePump'], Awaitable]
MAX_MESSAGE_ID = 99999
last_message_id = 0


def next_message_id() -> int:
    global last_message_id
    last_message_id += 1
    return last_message_id % MAX_MESSAGE_ID


class WebsocketConnection:
    next_id = 0

    def __init__(self,
                 connection: Optional[websockets.WebSocketCommonProtocol] = None,
                 request_handler: Optional[RequestHandlerType] = None,
                 close_handler: Optional[CloseHandlerType] = None,
                 message_queues: Optional[Dict[int, asyncio.Queue]] = None):
        self.id = WebsocketConnection.next_id
        WebsocketConnection.next_id += 1
        self.logger = logger.getChild(self.__class__.__name__)
        self.connection = connection
        self.request_handler = request_handler
        self.close_handler = close_handler
        self.event_loop = asyncio.get_event_loop()
        self.send_queue = asyncio.Queue()
        self.receive_queues = {} if message_queues is None else message_queues
        self.message_send_task: Optional[asyncio.Task] = None
        self.message_receive_task: Optional[asyncio.Task] = None
        self.running = False

    async def start(self, wait=True, connection: Optional[websockets.WebSocketCommonProtocol] = None):
        if connection is not None:
            self.connection = connection

        self.running = True
        self.message_send_task = asyncio.Task(self._message_receive())
        self.message_receive_task = asyncio.Task(self._message_send())

        if wait:
            await self.wait()

    async def wait(self):
        await self.message_send_task
        await self.message_receive_task

    async def stop(self):
        self.running = False
        await self.connection.close()
        await self.connection.wait_closed()
        self.message_send_task.cancel()
        self.message_receive_task.cancel()
        try:
            await self.message_receive_task
            await self.message_send_task
        except futures.CancelledError:
            pass

    def send_message_sync(self, message: messages.Message):
        utils.await_sync(self.send_message(message), self.event_loop)

    async def send_message(self, message: messages.Message):
        if self.connection is None:
            raise errors.NotConnectedError('Must be connected to send message: {message}')

        await self.send_queue.put(message)

    def send_request_sync(self, method: str, data: Any = None) -> messages.Response:
        return utils.await_sync(self.send_request(method, data), self.event_loop)

    async def send_request(self, method: str, data: Any = None) -> messages.Response:
        return await self.request(messages.Request(method=method, data=data))

    def send_response_sync(self, id: int, data: Any = None, status: int = messages.Status.SUCCESS):
        utils.await_sync(self.send_response(id, data, status), self.event_loop)

    async def send_response(self, id: int, data: Any = None, status: int = messages.Status.SUCCESS):
        response = messages.Response(id=id, status=status, data=data)
        await self.send_message(response)

    def send_error_response_sync(self, id: int, error: Exception):
        utils.await_sync(self.send_error_response(id, error), self.event_loop)

    async def send_error_response(self, id: int, error: Exception):
        response = messages.Response(id=id, status=messages.Status.ERROR, data=error)
        await self.send_message(response)

    def request_sync(self, request: messages.Request, raise_error=True) -> messages.Response:
        return utils.await_sync(self.request(request, raise_error))

    async def request(self, request: messages.Request, raise_error=True) -> messages.Response:
        request.id = next_message_id()
        queue = self.receive_queues[request.id] = asyncio.Queue()
        await self.send_message(request)

        try:
            response: messages.Response = await queue.get()
            self.receive_queues.pop(request.id)
        except asyncio.TimeoutError:
            self.receive_queues.pop(request.id)
            raise errors.RequestTimeoutError(f'Request timed out while waiting for response: {request}')

        if response.status == messages.Status.ERROR and raise_error:
            raise response.data

        return response

    async def _handle_request(self, request: messages.Request):
        id = request.id
        response = await self.request_handler(self, request)
        response.id = id
        await self.send_message(response)

    async def _message_send(self):
        while self.running:
            message: messages.Message = await self.send_queue.get()
            try:
                message_bytes = message.to_pickle()
            except (pickle.PickleError, AttributeError):
                message_bytes = messages.Response(id=message.id, status=messages.Status.ERROR, data=errors.SerializationError(f'Cannot serialize response data')).to_pickle()

            self.logger.info(f'[{self.id}] > {message}')
            await self.connection.send(message_bytes)

    async def _message_receive(self):
        try:
            while self.running:
                message_bytes = await self.connection.recv()
                try:
                    message = pickle.loads(message_bytes)
                except (pickle.UnpicklingError, TypeError):
                    self.logger.error(f'Invalid message: {message_bytes}')
                    continue

                self.logger.info(f'[{self.id}] < {message}')

                if isinstance(message, messages.Response):  # response
                    if message.id not in self.receive_queues:
                        raise errors.InvalidMessageIdError(f'No request found for response ID: {message.id}')

                    queue = self.receive_queues[message.id]
                    await queue.put(message)

                elif isinstance(message, messages.Request):  # request
                    if self.request_handler is not None:
                        asyncio.create_task(self._handle_request(message))

                else:
                    self.logger.error(f'Invalid message: {message}')

        except websockets.ConnectionClosed as err:
            self.logger.info(f'[{self.id}] Connection closed')

            if isinstance(err, websockets.ConnectionClosedError):
                self.logger.error(err)

            if self.close_handler is not None:
                await self.close_handler(self)


from typing import Optional, Awaitable, Any
import asyncio


def await_sync(awaitable: Awaitable, loop: Optional[asyncio.AbstractEventLoop] = None, timeout: Optional[float] = None) -> Any:
    loop = loop or asyncio.get_event_loop()
    if loop.is_running():
        future = asyncio.run_coroutine_threadsafe(awaitable, loop or asyncio.get_event_loop())
        return future.result(timeout=timeout)
    else:
        return loop.run_until_complete(awaitable)



from typing import Optional, List, Dict
import logging
import time
import asyncio
import websockets
from rpc_gateway import messages, errors
from rpc_gateway.websocket_connection import WebsocketConnection
from rpc_gateway.utils import await_sync

logger = logging.getLogger(__name__)


class GatewayConnection:
    def __init__(self, gateway: 'Gateway', connection: websockets.WebSocketServerProtocol):
        self.gateway = gateway
        self.connection = connection
        self.message_pump = WebsocketConnection(connection, lambda message: gateway.handle_request(self, message))
        self.instances: List[str] = []

    async def start(self):
        await self.message_pump.start()

    async def stop(self):
        await self.message_pump.stop()

    def register_instances(self, instances):
        self.instances = list(set(*self.instances, *instances))


class Gateway:
    SERVER_MESSAGES = (messages.Method.GET, messages.Method.SET, messages.Method.CALL)

    def __init__(self, host: str = 'localhost', port: int = 8888):
        self.host = host
        self.port = port
        self.logger = logger.getChild(self.__class__.__name__)
        self.websocket_connections: List[WebsocketConnection] = []
        self.websocket: Optional[websockets.WebSocketServer] = None
        self.event_loop = asyncio.get_event_loop()
        self.instances: Dict[str, WebsocketConnection] = {}
        self.websocket_instances: Dict[WebsocketConnection, List[str]] = {}

    def start(self, wait = True):
        await_sync(self._start(wait), self.event_loop)

    async def _start(self, wait = True):
        self.logger.info(f'Starting on ws://{self.host}:{self.port}')
        self.websocket = await websockets.serve(self.on_connection, self.host, self.port)

        if wait:
            await self._wait()

    def wait(self):
        await_sync(self._wait(), self.event_loop)

    async def _wait(self):
        if self.websocket is not None:
            await self.websocket.wait_closed()

        self.logger.info(f'Done')

    def stop(self):
        await_sync(self._stop(), self.event_loop)

    async def _stop(self):
        await asyncio.gather(*[websocket_connection.stop() for w, websocket_connection in self.websocket_connections.items()])
        self.websocket.close()

    async def on_connection(self, connection: websockets.WebSocketServerProtocol, path: str):
        self.logger.info(f'New connection from {connection.remote_address} path: {path}')
        websocket_connection = WebsocketConnection(connection, request_handler=self.handle_request, close_handler=self.handle_close)
        self.websocket_connections.append(websocket_connection)

        await websocket_connection.start()

        for instance in self.websocket_instances[websocket_connection]:
            self.instances.pop(instance)
        self.websocket_connections.remove(websocket_connection)

        self.logger.info(f'Connection from {connection.remote_address} closed')

    #
    # Request Handlers
    #

    async def handle_forward_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
        self.logger.info(f'Forwarding request to server: {request}')
        instance = request.data['instance']

        if instance not in self.instances:
            return messages.Response(status=messages.Status.ERROR, data=errors.InstanceNotFoundError(f'Instance not found: {instance}'))

        server = self.instances[request.data['instance']]
        response = await server.request(request, raise_error=False)
        self.logger.info(f'Forwarding response to client: {response}')

        return response

    async def handle_available_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
        available = request.data in self.instances
        return messages.Response(data=available)

    async def handle_register_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
        self.logger.info(f'Registering instances: {request.data}')
        self.websocket_instances[websocket_connection] = self.websocket_instances.get(websocket_connection, []) + request.data
        for instance_name in request.data:
            self.instances[instance_name] = websocket_connection

        return messages.Response()

    # this is called by the GatewayConnection MessagePump when a new request is received
    async def handle_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
        if request.method in self.SERVER_MESSAGES:
            return await self.handle_forward_request(websocket_connection, request)

        if request.method == messages.Method.AVAILABLE:
            return await self.handle_available_request(websocket_connection, request)

        if request.method == messages.Method.REGISTER:
            return await self.handle_register_request(websocket_connection, request)

        return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMethodError(f'Invalid method: {request.method}'))

    async def handle_close(self, websocket_connection: WebsocketConnection):
        if websocket_connection in self.websocket_instances:
            # remove registered instances
            for instance_name in self.websocket_instances[websocket_connection]:
                self.instances.pop(instance_name)

            self.websocket_instances.pop(websocket_connection)

        # send an error response for any in-progress requests
        for request_id, response_queue in websocket_connection.receive_queues.items():
            await response_queue.put(messages.Response(status=messages.Status.ERROR, data=errors.ServerConnectionLostError(f'Server connection lost')))


class GatewayClient:
    def __init__(self, gateway_url: str = 'ws://localhost:8888'):
        self.logger = logger.getChild(self.__class__.__name__)
        self.gateway_url = gateway_url
        self.websocket_connection = WebsocketConnection(request_handler=self._handle_request, close_handler=self._handle_close)
        self.event_loop = asyncio.get_event_loop()
        self.connect_retry_timeout = 2.0

    @property
    def connected(self) -> bool:
        return self.websocket_connection.connection is not None

    async def _connect(self):
        while True:
            try:
                self.logger.info(f'Connecting to {self.gateway_url}')
                self.connection = await websockets.connect(self.gateway_url)
                return
            except OSError:
                self.logger.warning(f'Error connecting to {self.gateway_url}, retrying in {self.connect_retry_timeout} seconds')
                time.sleep(self.connect_retry_timeout)

    def start(self, wait=True):
        await_sync(self._start(wait), self.event_loop)

    async def _start(self, wait=True):
        await self._connect()
        await self.websocket_connection.start(wait=False, connection=self.connection)
        await self._on_start()

        if wait:
            await self._wait()

    def wait(self):
        await_sync(self._wait(), self.event_loop)

    async def _wait(self):
        await self.websocket_connection.wait()

    async def _on_start(self):
        pass

    def stop(self):
        await_sync(self._stop(), self.event_loop)

    async def _stop(self):
        await self.websocket_connection.stop()

    async def _handle_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
        pass

    async def _handle_close(self, websocket_connection: WebsocketConnection):
        await self._connect()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    gateway = Gateway()
    gateway.start()


from typing import Any, Iterable, Optional, List, Dict, Callable
import time
import logging
import asyncio
from rpc_gateway import errors, messages, gateway


class Client(gateway.GatewayClient):
    def get_instance(self, instance_name) -> Any:
        if not self.connected:
            self.start(wait=False)
            time.sleep(1)

        if not self.instance_available(instance_name):
            raise errors.InstanceNotFoundError(f'Instance not found: {instance_name}')

        client = self

        class _Proxy:
            def __dir__(self) -> Iterable[str]:
                return client.call(instance_name, '__dir__')

            def __getattr__(self, item):
                return client.get(instance_name, item)

            def __setattr__(self, key, value):
                return client.set(instance_name, key, value)

            def __repr__(self):
                return client.call(instance_name, '__repr__')

            def __str__(self):
                return client.call(instance_name, '__str__')

        return _Proxy()

    def proxy_method(self, instance_name: str, method_name: str) -> Callable:
        def _proxy_method(*args, **kwargs):
            return self.call(instance_name, method_name, args, kwargs)

        return _proxy_method

    def instance_available(self, instance_name: str) -> bool:
        response = self.websocket_connection.send_request_sync(messages.Method.AVAILABLE, instance_name)
        return response.data

    def call(self, instance_name: str, method_name: str, args: Optional[Iterable[Any]] = None, kwargs: Optional[Dict[str, Any]] = None) -> Any:
        response = self.websocket_connection.send_request_sync(messages.Method.CALL, {
            'instance': instance_name,
            'attribute': method_name,
            'args': args or [],
            'kwargs': kwargs or {}
        })

        return response.data

    def get(self, instance_name: str, attribute_name: str) -> Any:
        response = self.websocket_connection.send_request_sync(messages.Method.GET, {
            'instance': instance_name,
            'attribute': attribute_name
        })

        if response.status == messages.Status.METHOD:
            return self.proxy_method(instance_name, attribute_name)

        return response.data

    def set(self, instance_name: str, attribute_name: str, value: Any) -> Any:
        self.websocket_connection.send_request_sync(messages.Method.SET, {
            'instance': instance_name,
            'attribute': attribute_name,
            'value': value
        })

        return value


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    class TestClass:
        foo = 'bar'

        def method(self):
            return 'baz'

    client = Client()
    test: TestClass = client.get_instance('test')
    print(test.foo)
    # print(test.method())
    # print(test)
